#include "DfpFlashModel.h"
#include "DfpFsmStates.h"
#include "DfpFlashModelColumns.h"

#include "base/DeviceIndexEnabled.h"
#include "device-modules/ModularDeviceStatus.h"

#include <QColor>

namespace {
static bool hasChanges(const ModularDeviceConfig &c1, const ModularDeviceConfig &c2)
{
    return c1.flashProg != c1.flashProg;
}

static bool hasChanges(const BaseDeviceAppConfig &c1, const BaseDeviceAppConfig &c2)
{
    for (auto it1=c1.knownConfigs.cbegin(); it1!=c1.knownConfigs.cend(); ++it1) {
        auto it2 = c2.knownConfigs.constFind(it1.key());
        if(it2 == c2.knownConfigs.cend())
            return true;
        if (hasChanges(*it1, *it2))
            return true;
    }
    return false;
}

} // namespace

struct DfpFlashModelPrivate
{
    int rowOf(const DeviceIndex &deviceIndex) const;
    DeviceIndex indexOf(int row) const;

    QMap<DeviceIndex, FsmState> fsmStates; // from DfpFsmDev
    BaseDeviceAppConfig config;
};

int DfpFlashModelPrivate::rowOf(const DeviceIndex &deviceIndex) const
{
    return std::distance(config.knownConfigs.begin(), config.knownConfigs.find(deviceIndex));
}

DeviceIndex DfpFlashModelPrivate::indexOf(int row) const
{
    int r = 0;
    for (auto it = config.knownConfigs.begin(); it != config.knownConfigs.end(); ++it)
        if (r++ == row)
            return it.key();
    return {};
}

DfpFlashModel::DfpFlashModel(QObject *parent)
    : DeviceIndexTableModel(parent)
    , d_ptr(new DfpFlashModelPrivate)
{
    setObjectName("Flash image");
}

DfpFlashModel::~DfpFlashModel()
{
    delete d_ptr;
}

int DfpFlashModel::rowCount(const QModelIndex &) const
{
    Q_D(const DfpFlashModel);
    return d->config.deviceKeys().size();
}

int DfpFlashModel::columnCount(const QModelIndex &) const
{
    return int(DfpFlashModelColumns::COL_NUMBER);
}

DeviceIndex DfpFlashModel::deviceIndexOf(int row) const
{
    Q_D(const DfpFlashModel);
    return d->indexOf(row);
}

int DfpFlashModel::rowOf(const DeviceIndex &sourceIndex) const
{
    Q_D(const DfpFlashModel);
    return d->rowOf(sourceIndex);
}

QVariant DfpFlashModel::data(const QModelIndex &index, int role) const
{
    auto deviceIndex = deviceIndexOf(index.row());
    if (deviceIndex.isNull())
        return QVariant();

    auto col = DfpFlashModelColumns(index.column());
    switch (role) {
    case Qt::DisplayRole: return data_DisplayRole(deviceIndex, col);
    case Qt::EditRole: return data_EditRole(deviceIndex, col);
    case Qt::ToolTipRole: return data_ToolTip(deviceIndex, col);
    case Qt::BackgroundRole: return data_BackgroundRole(deviceIndex, col);
    }
    return QVariant();
}

QVariant DfpFlashModel::headerData(int section, Qt::Orientation orientation, int role) const
{
    if (orientation == Qt::Horizontal && role == Qt::DisplayRole) {
        switch (DfpFlashModelColumns(section)) {
        case DfpFlashModelColumns::COL_FILE_NAME: return "File";
        case DfpFlashModelColumns::COL_BROWSE_PATH: return "Browse path";
        default:;
        }
    }
    return DeviceIndexTableModel::headerData(section, orientation, role);
}

bool DfpFlashModel::setData(const QModelIndex &index, const QVariant &value, int role)
{
    Q_D(DfpFlashModel);
    const auto &deviceIndex = deviceIndexOf(index.row());
    if (deviceIndex.isNull())
        return {};

    const auto it = d->config.knownConfigs.find(deviceIndex);
    if (it == d->config.knownConfigs.end())
        return false;

    if (role == Qt::EditRole) {
        switch (DfpFlashModelColumns(index.column())) {
        case DfpFlashModelColumns::COL_BROWSE_PATH:
            if(it->flashProg.imageFilePath == value.toString())
                return true;
            it->flashProg.imageFilePath = value.toString();
            emit dataChanged(index.siblingAtColumn(int(DfpFlashModelColumns::COL_FILE_NAME)),
                             index,
                             {Qt::DisplayRole, Qt::BackgroundRole});
            emit configChanged(d->config);
            return true;
        default:;
        }
    }
    return false;
}

void DfpFlashModel::setDevState(const DeviceIndex &devIndex, FsmState fsmSt)
{
    Q_D(DfpFlashModel);
    auto it = d->fsmStates.find(devIndex);
    if(it == d->fsmStates.end() || *it == fsmSt)
        return;
    *it = fsmSt;
    auto row = d->rowOf(devIndex);
    auto i = index(row, int(DfpFlashModelColumns::COL_FILE_NAME));
    emit dataChanged(i, i, {Qt::BackgroundRole});
}

void DfpFlashModel::setConfig(const ConfigKey &, const BaseDeviceAppConfig &newConfig)
{
    Q_D(DfpFlashModel);
    if (d->config == newConfig)
        return;

    bool performReset = d->config.deviceKeys() != newConfig.deviceKeys();
    bool hasAnyChanges = performReset ? true : hasChanges(d->config, newConfig);

    if (performReset) {
        beginResetModel();
        const auto &newSet = newConfig.deviceIndexSet();
        const auto &oldSet = d->config.deviceIndexSet();
        d->config = newConfig;
        const auto toRemove = oldSet-newSet;
        for (const auto &i : toRemove)
            d->fsmStates.remove(i);
        endResetModel();
    } else {
        d->config = newConfig;
        if(hasAnyChanges)
            sendDataChanged();
    }
}

void DfpFlashModel::sendDataChanged(QVector<int> roles, int row, int col)
{
    int top, bottom, left, right;
    if(row == -1) {
        top = 0;
        bottom = rowCount()-1;
    } else {
        top=bottom=row;
    }
    if(col == -1) {
        left = 0;
        right = columnCount()-1;
    } else {
        left=right=col;
    }
    if(roles.isEmpty())
        roles = {Qt::DisplayRole, Qt::BackgroundRole};
    emit dataChanged(index(top,left), index(bottom, right), roles);
}

QVariant DfpFlashModel::data_DisplayRole(const DeviceIndex &index, DfpFlashModelColumns column) const
{
    switch (column) {
    case DfpFlashModelColumns::COL_FILE_NAME:
    {
        auto path = getImagePath(index).split("/").last();
        return path.isEmpty() ? "--- missing ---" : path;
    }
    default:;
    }
    return {};
}

QVariant DfpFlashModel::data_EditRole(const DeviceIndex &index, DfpFlashModelColumns column) const
{
    switch (column) {
    case DfpFlashModelColumns::COL_BROWSE_PATH:
        return getImagePath(index);
    default:;
    }
    return {};
}

QVariant DfpFlashModel::data_ToolTip(const DeviceIndex &index, DfpFlashModelColumns column) const
{
    switch (column) {
    case DfpFlashModelColumns::COL_FILE_NAME:
        return getImagePath(index);
    default:;
    }
    return {};
}

QVariant DfpFlashModel::data_BackgroundRole(const DeviceIndex &index, DfpFlashModelColumns column) const
{
    Q_D(const DfpFlashModel);
    switch (column) {
    case DfpFlashModelColumns::COL_FILE_NAME:
    {
        auto it = d->fsmStates.constFind(index);
        if(it == d->fsmStates.cend())
            return {};

        switch (*it) {
        case DfpFsmStateError: return QColor(Qt::red);
        case DfpFsmStateMine: return QColor(Qt::darkGreen);
        case DfpFsmStateAlien: return QColor(Qt::darkYellow);
        default: return {};
        }
    }
    default:;
    }
    return {};
}

QString DfpFlashModel::getImagePath(const DeviceIndex &index) const
{
    Q_D(const DfpFlashModel);
    return d->config.knownConfigs.value(index).flashProg.imageFilePath;
}

QMap<int, bool> DfpFlashModel::defaultColumnState() const
{
    return  {
        {int(DfpFlashModelColumns::COL_FILE_NAME), true},
        {int(DfpFlashModelColumns::COL_BROWSE_PATH), true},
    };

}
