/*
**    Copyright 2017-2022 Ivan Filippov, Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
**
*/

#include "DfpMainWindow.h"
#include "ui_DfpMainWindow.h"

#include <QDateTime>
#include <QDebug>
#include <QElapsedTimer>
#include <QLabel>
#include <QMessageBox>
#include <QSettings>
#include <QTimer>

#include "DfpFsmMain.h"
#include "DfpFsmStates.h"
#include "DfpTask.h"
#include "DfpSignalRouter.h"
#include "EthFlashDev.h"
#include "EthFlashDevGlobal.h"

#include "base/fw_version.h"
#include "flash-prog/FpgaDataBase.h"
#include "fsm/FsmState.h"
#include "lib-common/ReadoutAppConfigDb.h"
#include "lib-common/ReadoutAppConfigDispatcher.h"
#include "lib-common/ReadoutAppSignalRouter.h"
#include "mldiscover/DeviceDispatcher.h"
#include "widgets/FileDialog.h"

namespace {
constexpr fw_version_t REV_ETHERNET_WRITE = fw_version_t(1, 0, 24200);
constexpr const auto defaultSearchPath = "/net/afi-nas/vol/firmware/fpga/auto";
QString getSearchPath()
{
    const QSettings settings;
    return settings.value("searchPath", defaultSearchPath).toString();
}

void setSearchPath(const QString &path)
{
    QSettings settings;
    settings.setValue("searchPath", path);
}

bool getSingleCheck()
{
    const QSettings settings;
    return settings.value("singleCheck", false).toBool();
}

} // namespace

Q_DECLARE_METATYPE(fw_version_t)
DfpMainWindow::DfpMainWindow(QWidget *parent)
    : ReadoutAppMainWindow(parent)
    , ui(new Ui::DfpMainWindow)
    , configDb(new ReadoutAppConfigDb(this))
    , fsm(new DfpFsmMain(this))
    , deviceDispatcher(new DeviceDispatcher(this))
    , lastStatusLabel(new QLabel(this))
    , dataBase(new FpgaDataBase(this))
    , singleCheck(getSingleCheck())
{
    qRegisterMetaType<DFP_TASK>();
    qRegisterMetaType<FlashImageType>();

    ui->setupUi(this);
    setupMenu(ui->menuFile, ui->menuHelp, ui->menuOptions, ui->menuTools);

    ui->mainToolBar->hide();

    auto configDispatcher = new ReadoutAppConfigDispatcher(this);
    configDispatcher->subscribeConfig(configDb);
    configDispatcher->subscribeCS(configDb);

    auto signalRouter = &DfpSignalRouter::instance();
    signalRouter->setReadoutAppMainWindow(this);
    signalRouter->setConfigDispatcher(configDispatcher);
    signalRouter->setDeviceDispatcher(deviceDispatcher);
    signalRouter->setDeviceMgrFsm(fsm);
    signalRouter->setDfpStatusModel(ui->deviceStatusTableView);
    signalRouter->connectAll();

    connect(ui->pushButtonSelectDevices, &QPushButton::clicked,
            this, &ReadoutAppMainWindow::actionDiscoverDialog);

    deviceDispatcher->subscribeDevices(configDispatcher);

    connect(fsm, &BaseFsm::stateChanged,
            this, &DfpMainWindow::onFsmStateChange);
    connect(this, &DfpMainWindow::printFsmInput,
            fsm, &BaseFsm::printFsmInput);

    ui->actionSingle_check_mode->setChecked(singleCheck);

    statusBar()->addPermanentWidget(lastStatusLabel);

    ui->verticalLayoutCentral->setAlignment(ui->horizontalLayoutButtons, Qt::AlignTop);
    ui->deviceStatusTableView->setKeepDisabledDev(true);
}

DfpMainWindow::~DfpMainWindow()
{
    delete ui;
}

DeviceStatusTableViewFancy *DfpMainWindow::getDeviceStatusTableView()
{
    return ui->deviceStatusTableView;
}

void DfpMainWindow::on_actionWrite_Golden_Image_triggered()
{
    QMessageBox msgBox;
    msgBox.setText("Do you want to write \"Golden Image\"?");
    msgBox.setStandardButtons(QMessageBox::Yes | QMessageBox::No);
    msgBox.setDefaultButton(QMessageBox::No);
    int ret = msgBox.exec();
    if (ret != QMessageBox::Yes)
        return;
    if (!alienStateCheck())
        return;

    curTask = {DFP_TASK_WRITE, FlashImageType::Golden};
    fsm->setRequestState(DfpFsmStateWriting);
}

void DfpMainWindow::on_actionGet_record_logs_triggered()
{
    auto dis = getDeviceStatusTableView()->getSelectedDeviceIndexes();

    for (const auto &i : dis) {
        qInfo().noquote() << QString("\t=== Log for %1 ===").arg(i.getSerialStr());
        dataBase->readInfo(i.getSerial());
    }
}

void DfpMainWindow::on_actionVerify_Golden_image_triggered()
{
    curTask = {DFP_TASK_VERIFY, FlashImageType::Golden};
    fsm->setRequestState(DfpFsmStateReading);
}

void DfpMainWindow::on_actionVerify_regular_image_triggered()
{
    curTask = {DFP_TASK_VERIFY, FlashImageType::Regular};
    fsm->setRequestState(DfpFsmStateReading);
}

void DfpMainWindow::on_actionErase_Golden_part_triggered()
{
    startErase(FlashImageType::Golden);
}

void DfpMainWindow::on_actionErase_regular_part_triggered()
{
    startErase(FlashImageType::Regular);
}

void DfpMainWindow::on_pushButtonWrite_clicked()
{
    if (!alienStateCheck())
        return;

    curTask = {DFP_TASK_WRITE, FlashImageType::Regular};
    fsm->setRequestState(DfpFsmStateWriting);
}

void DfpMainWindow::deviceConnected(const DeviceIndex &index, QString fwStr, const fw_version_t &fwVersion)
{
    // const int row = getRowByIndex(index);
    // if (row != -1) {
    //     const bool supported = fwVersion >= REV_ETHERNET_WRITE;
    //     ui->tableWidget->item(row, COL_FIRMWARE)->setText(fwStr);
    //     ui->tableWidget->item(row, COL_FIRMWARE)->setTextColor(supported ? Qt::black : Qt::red);
    //     updateTableSize();
    // }
}

void DfpMainWindow::imageWritten(const DeviceIndex &index, FlashImageType imageType, const QString &path)
{
    if (dataBase)
        dataBase->writeInfo(index.getSerial(), imageType == FlashImageType::Golden, path);
}

void DfpMainWindow::liveMagicFailed(const DeviceIndex &index)
{
    auto cfg = config;
    auto it = cfg.knownConfigs.find(index);
    if(it == cfg.knownConfigs.end())
        return;
    it->enabled = false;
    emit configChanged(cfg);
}

void DfpMainWindow::onFsmStateChange(const FsmState &state, const FsmState &oldState)
{
    ReadoutAppMainWindow::onFsmStateChange(state);

    switch (state) {
    case DfpFsmStateErasing:
    case DfpFsmStateReading:
    case DfpFsmStateWriting: {
        taskTimer.start();
        Q_ASSERT(curTask.has_value());
        const auto task = curTask->type;
        const auto imageType = curTask->isGold;
        setLabelStatus("in progress...", "gray");

        QString taskStr = dfpTaskToString(task);
        qInfo().noquote() << QString("\n=== Start %1 ===").arg(taskStr);
        qInfo().noquote() << QString("%1 %2 image %3")
                                 .arg(QDateTime::currentDateTime().toString())
                                 .arg(flashImageTypeStr(imageType))
                                 .arg(taskStr);

        emit startDevTask(task, imageType);
        break;
    }
    case DfpFsmStateDone:
        emit allowFsmTransit();
        break;
    default:;
    }

    switch (oldState) {
    case DfpFsmStateErasing:
    case DfpFsmStateReading:
    case DfpFsmStateWriting: {
        curTask.reset();

        bool allOk = state == DfpFsmStateDone;

        qInfo().noquote() << "===" << (allOk ? "end" : "fail") << "===";
        setLabelStatus(QString("%1 (%2)")
                           .arg(allOk ? QString("ok in %1 sec").arg(taskTimer.elapsed() / 1000) : "fail")
                           .arg(QTime::currentTime().toString()),
                       allOk ? "green" : "red");
        break;
    }
    default:;
    }

    setPushButtonEnabled();
}

void DfpMainWindow::devFileUpdated(const DeviceIndex &index, const QString &filePath)
{
    auto cfgIt = config.knownConfigs.find(index);
    if (cfgIt != config.knownConfigs.end() && cfgIt->flashProg.imageFilePath != filePath) {
        if(cfgIt->flashProg.imageFilePath == filePath)
            return;
        cfgIt->flashProg.imageFilePath = filePath;
        saveConfig();
    }
}

void DfpMainWindow::setConfig(const ConfigKey &configKey, const BaseDeviceAppConfig &newConfig)
{
    if (config == newConfig)
        return;
    const auto &oldEnSet = config.deviceIndexSet(true);
    const auto &newEnSet = newConfig.deviceIndexSet(true);
    ReadoutAppMainWindow::setConfig(configKey, newConfig);

    const auto toRemoveDevice = oldEnSet - newEnSet;
    for (const auto &i : toRemoveDevice)
        deleteDevice(i);

    const auto toAddDevice = newEnSet - oldEnSet;
    for (const auto &i : toAddDevice)
        addDevice(i);
}

void DfpMainWindow::startErase(FlashImageType imageType)
{
    QMessageBox msgBox(this);
    msgBox.setText(
        QString("Do you want to erase \"%1 Image\"?").arg(flashImageTypeStr(imageType)));
    msgBox.setStandardButtons(QMessageBox::Yes | QMessageBox::No);
    msgBox.setDefaultButton(QMessageBox::No);
    int ret = msgBox.exec();
    if (ret == QMessageBox::Yes) {
        curTask = {DFP_TASK_ERASE, imageType};
        fsm->setRequestState(DfpFsmStateErasing);
    }
}

void DfpMainWindow::setPushButtonEnabled() const
{
    auto st = fsm->getState();
    bool isReady = st == DfpFsmStateMine || st == DfpFsmStateAlien;
    bool isError = st == DfpFsmStateError;

    ui->deviceStatusTableView->setControlEnabled(isReady || isError);

    ui->pushButtonSelectDevices->setEnabled(isReady || isError);
    ui->pushButtonRefreshPaths->setEnabled(isReady || isError);
    for(auto a : {ui->actionSingle_check_mode,
        getActionDiscoverDialog(),
        getActionConfigurationManger()}) {
        a->setEnabled(isReady || isError);
    }

    for(auto a : {ui->actionVerify_Golden_image,
        ui->actionVerify_regular_image,
        ui->actionErase_Golden_part,
        ui->actionErase_regular_part}) {
        a->setEnabled(isReady);
    }
    if (isReady) {
        updateWriteButtonState();
    } else {
        ui->pushButtonWrite->setEnabled(isReady);
        ui->actionWrite_Golden_Image->setEnabled(isReady);
    }
}

void DfpMainWindow::setLabelStatus(QString msg, QString color) const
{
    lastStatusLabel->setText(msg);
    lastStatusLabel->setStyleSheet(QString("QLabel { color : %1; }").arg(color));
}

bool DfpMainWindow::alienStateCheck()
{
    if (fsm->getState() == DfpFsmStateAlien) {
        QMessageBox msgBox(this);
        msgBox.setText("Some devices will wrote image with non-native signature\n"
                       "Proceed?");
        msgBox.setStandardButtons(QMessageBox::Yes | QMessageBox::No);
        msgBox.setDefaultButton(QMessageBox::No);
        int ret = msgBox.exec();
        return ret == QMessageBox::Yes;
    }
    return true;
}

void DfpMainWindow::addDevice(const DeviceIndex &index)
{
    Q_ASSERT(!devices.contains(index));
    auto devGlob = new EthFlashDevGlobal(index, this);
    devices.insert(index, QSharedPointer<EthFlashDevGlobal>(devGlob));
    const auto &dev = devGlob->getDevice<EthFlashDev>();

    qRegisterMetaType<fw_version_t>();
    connect(dev, &EthFlashDev::fsmStateChanged,
            fsm, &DfpFsmMain::setDevState);

    connect(dev, &EthFlashDev::devConnected,
            this, &DfpMainWindow::deviceConnected);
    connect(dev, &EthFlashDev::liveMagicFailed,
            this, &DfpMainWindow::liveMagicFailed);
    connect(dev, &EthFlashDev::imageWritten,
            this, &DfpMainWindow::imageWritten);

    connect(this, &DfpMainWindow::printFsmInput,
            dev, &EthFlashDev::printFsmInput);
    connect(this, &DfpMainWindow::allowFsmTransit,
            dev, &EthFlashDev::allowFsmTransit);

    connect(this, &DfpMainWindow::setSingleCheckMode,
            dev, &EthFlashDev::setSingleCheckMode);
    connect(this, &DfpMainWindow::autoDetectFile,
            dev, &EthFlashDev::autoDetectFile);
    connect(this, &DfpMainWindow::startDevTask,
            dev, &EthFlashDev::startTask);
    connect(dev, &EthFlashDev::devFileChanged,
            this, &DfpMainWindow::devFileUpdated);

    connect(dev, &ModularMregDevice::statusUpdated,
            this, &ReadoutAppMainWindow::setStatus);

    auto signalRouter = &DfpSignalRouter::instance();
    signalRouter->connect_device(devGlob, dev);
    deviceDispatcher->subscribeHostAddressUpdates(index, dev);
    emit setSingleCheckMode(singleCheck);

    qInfo() << QString("Added device %1").arg(index.getIdent());
}

void DfpMainWindow::deleteDevice(const DeviceIndex &i)
{
    if (i.isNull()) {
        devices.clear();
    } else {
        devices.remove(i);
    }
}

void DfpMainWindow::saveConfig()
{
    emit configChanged(config);
}

void DfpMainWindow::updateWriteButtonState() const
{
    bool enable = fsm->getState() == DfpFsmStateMine || fsm->getState() == DfpFsmStateAlien;
    //    enable &= hasEnabledDevice;

    ui->pushButtonWrite->setEnabled(enable);
    ui->actionWrite_Golden_Image->setEnabled(enable);
}

void DfpMainWindow::on_pushButtonRefreshPaths_clicked()
{
    emit autoDetectFile(DeviceIndex(), getSearchPath());
}

void DfpMainWindow::on_actionPrint_FSM_input_triggered() const
{
    emit printFsmInput();
}

void DfpMainWindow::on_actionSet_common_bit_file_triggered()
{
    const constexpr static char* KEY = "commonFile";
    QSettings settings;
    auto fPath = settings.value(KEY).toString();

    fPath = FileDialog::getOpenFileName(this,
                                        tr("Open Bit file"),
                                        fPath,
                                        tr("Bit Files (*.bit)"));
    if(fPath.isEmpty())
        return;

    settings.setValue(KEY, fPath);
    for(auto i=config.knownConfigs.begin(); i!=config.knownConfigs.end(); ++i) {
        if(!i->enabled)
            continue;
        i->flashProg.imageFilePath = fPath;
    }

    saveConfig();
}

void DfpMainWindow::on_actionSet_auto_detect_folder_triggered()
{
    QString dir = FileDialog::getExistingDirectory(this, tr("Open FPGA directory"), getSearchPath());
    if (dir.isEmpty())
        return;
    setSearchPath(dir);
    saveConfig();
}

void DfpMainWindow::on_actionSingle_check_mode_toggled(bool checked)
{
    if(singleCheck == checked)
        return;
    singleCheck = checked;
    QSettings settings;
    settings.setValue("singleCheck", singleCheck);

    emit setSingleCheckMode(singleCheck);
}
