#include "MscInputModule.h"

#include "MscInputConfig.h"
#include "regio/regop.h"
#include "sdb/AfiSdbId.h"

namespace {
enum {
    // Ver 1.0
    REG_CTRL = 0x0,
    REG_CURRENT_STATE_OF_EXTERNAL_CE    = 0x1,
    REG_CURRENT_STATE_OF_INPUT_HITS     = 0x2,

    REG_NUMBER_OF_INPUT_HIT_CHANNELS    = 0xD,
    REG_NUMBER_OF_HISTOGRAMMERS         = 0xE,
    REG_NUMBER_OF_EXTERNAL_CE_INPUTS    = 0xF,

    // Ver 2.0
    REG_ID                              = 0x0,
    // REG_COTROL                          = 0x2, // unused
    REG_LOCK                            = 0x3,
    REG_V2_CURRENT_STATE_OF_EXTERNAL_CE = 0x4,
    REG_V2_CURRENT_STATE_OF_INPUT_HITS  = 0x6,
    REG_POLARITY                        = 0x8,

    REG_CE_EDGE_TIME_BASE = 0x80,

    REG_HIST_CH_SOURCE_BASE = 0xC0,

    REG_GATE_LUT_BASE = 0x100,
    REG_HIST_CE_LUT_BASE = 0x180,
};

enum {
    REG_CTRL_BIT_CNT_LOCK = 1 << 1,
};

const quint16 REG_LOCK_BIT_LOCK = 1;
} // namespace


QList<AFI_SDB_ID> MscInputModule::getSupportedSdbIds()
{
    return {AFI_SDB_MSC_INPUT};
}

void MscInputModule::post_init()
{
    BaseDeviceModule::post_init();

    readBuildParam();
}

bool MscInputModule::writeConfig(const MscInputConfig &cfg)
{
    auto r = setConfigRegOp(cfg);
    return regOpExecRebased(r);
}

bool MscInputModule::readStatus(MscInputStatus *result)
{
    auto r = getStatusRegOp(result);
    return regOpExecRebased(r);
}

std::optional<MscInputStatus> MscInputModule::readStatus()
{
    MscInputStatus status;
    if (!readStatus(&status))
        return {};
    return status;
}

QVector<quint64> MscInputModule::readCmpState(int num_samples)
{
    QVector<quint16> cmpState(num_samples);
    RegOpVector r;
    quint16 regSt = 0;
    switch (majorVersion()) {
    case 1: regSt = REG_CURRENT_STATE_OF_INPUT_HITS; break;
    case 2: regSt = REG_V2_CURRENT_STATE_OF_INPUT_HITS; break;
    default:
        return {};
    }
    for (int i = 0; i < num_samples; ++i)

        r.RegRead16(regSt, &cmpState[i]); // read multiple times
    if (!regOpExecRebased(r))
        return {};
    QVector<quint64> result;
    for (int i = 0; i < num_samples; ++i)
        result.push_back(cmpState[i]);
    return result;
}

void MscInputModule::addCounterLock(mlink::RegOpVector *r, bool lock) const
{
    switch (majorVersion()) {
    case 1:
        r->RegWrite16(REG_CTRL, lock ? REG_CTRL_BIT_CNT_LOCK : 0);
        return;
    case 2:
        r->RegWrite16(REG_LOCK, lock ? REG_LOCK_BIT_LOCK : 0);
        return;
    default:;
    }
}

void MscInputModule::readBuildParam()
{
    buildParams.reset();

    quint32 id=0;
    MscInputBuildParams st;
    RegOpVector r;
    switch (majorVersion()) {
    case 2:
        r.RegRead32(REG_ID, &id);
        [[fallthrough]];
    case 1:
        r.RegRead16(REG_NUMBER_OF_INPUT_HIT_CHANNELS, &st.channels);
        r.RegRead16(REG_NUMBER_OF_HISTOGRAMMERS, &st.hists);
        r.RegRead16(REG_NUMBER_OF_EXTERNAL_CE_INPUTS, &st.ceInputs);
        break;
    default:
        return;
    }

    if (!regOpExecRebased(r))
        return;

    if(majorVersion()==1 || getSupportedSdbIds().contains(static_cast<AFI_SDB_ID>(id)))
        buildParams = st;
}

int MscInputModule::getCeInputCount() const
{
    if (!buildParams.has_value())
        return 0;
    return buildParams->ceInputs;
}

mlink::RegOpVector MscInputModule::setConfigRegOp(const MscInputConfig &cfg) const
{
    RegOpVector r;
    if (!buildParams)
        return {};

    int lutSizeBit = 1 << buildParams->ceInputs;
    quint16 lutSizeWord16 = lutSizeBit / 16;
    lutSizeWord16 = std::max(lutSizeWord16, quint16(1));

    quint16 polarityMask = 0;
    for (uint16_t hId = 0; hId < buildParams->hists; ++hId) {
        if(cfg[hId].isActiveHigh)
            polarityMask |= 1<<hId;

        auto chId = cfg[hId].histCh;
        r.RegWrite16(REG_HIST_CH_SOURCE_BASE + hId, chId);

        quint16 in_addr = REG_GATE_LUT_BASE + hId * lutSizeWord16;
        r.RegWrite16(in_addr, cfg[hId].histLut);

        in_addr = REG_HIST_CE_LUT_BASE + hId * lutSizeWord16;
        r.RegWrite16(in_addr, cfg[hId].ceLut);
    }
    if(majorVersion() == 2)
        r.RegWrite16(REG_POLARITY, polarityMask);
    return r;
}

RegOpVector MscInputModule::getStatusRegOp(MscInputStatus *result) const
{
    auto ceInputCount = getCeInputCount();
    result->ceInputCount = ceInputCount;
    result->ceInputEdgeTimesRaw.resize(ceInputCount);
    RegOpVector r;
    addCounterLock(&r, true); // lock counters
    r.RegWrite16(REG_CTRL, REG_CTRL_BIT_CNT_LOCK); // lock counters
    switch (majorVersion()) {
    case 1:
        r.RegRead16(REG_CURRENT_STATE_OF_EXTERNAL_CE, &result->ceInputStateRaw);
        r.RegRead16(REG_CURRENT_STATE_OF_INPUT_HITS, &result->hitState);
        break;
    case 2:
        r.RegRead16(REG_V2_CURRENT_STATE_OF_EXTERNAL_CE, &result->ceInputStateRaw);
        r.RegRead16(REG_V2_CURRENT_STATE_OF_INPUT_HITS, &result->hitState);
        break;
    default:;
    }
    for (uint16_t hId = 0; hId < ceInputCount; ++hId) {
        uint16_t addr = REG_CE_EDGE_TIME_BASE + hId * 8;
        r.RegRead64(addr, &result->ceInputEdgeTimesRaw[hId].first);     // rising
        r.RegRead64(addr + 4, &result->ceInputEdgeTimesRaw[hId].second); // falling
    }
    addCounterLock(&r, false); // lock counters
    return r;
}
